/*******************************************************************************/
/* Cruise controller                                                           */
/* Author: Peter Schrammel                                                     */
/*******************************************************************************/
/* modelled after                                                              */
/*     Robert Bosch GmbH: Bosch Automotive Handbook. Bentley (2007)            */
/* if you reuse this code, please cite also                                    */
/*      Peter Schrammel, Tom Melham, Daniel Kroening. Chaining Test Cases for 
        Reactive System Testing. In Testing Software and Systems
        LNCS Volume 8254, 2013, pp 133-148                                     */
/*******************************************************************************/

// uncomment this to compile the harness
#define CBMC

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

typedef struct state {
  int mode;
  int speed;
  int button;
} t_state;

#define I_BUTTON 1
#define I_ACC 2
#define I_DEC 3
#define I_GAS 4
#define I_BRAKE 5

typedef struct input {
  int signal; 
} t_input;

#ifndef CBMC
#include "cruise_tests.h"
#define TESTVECTOR testvector0
#define TESTLENGTH testvector0_length
#endif

void init(t_state *s) {
  s->mode = 0;
  s->speed = 0;
  s->button = 0;
}

void compute(t_input* i, t_state *s) {
  if((s->mode==0) && (s->speed==0) && (s->button==0)) {
    if((i->signal==I_ACC)||(i->signal==I_GAS)) s->speed=1;
    else if(i->signal==I_BUTTON) s->button=1;
  }
  else if((s->mode==0) && (s->speed==1) && (s->button==0)) {
    if((i->signal==I_BRAKE)||(i->signal==I_DEC)) s->speed=0;
    else if((i->signal==I_GAS)||(i->signal==I_ACC)) s->speed=2;
    else if(i->signal==I_BUTTON) { s->button=1; s->mode=1; }
  }
  else if((s->mode==0) && (s->speed==0) && (s->button==1)) {
    if((i->signal==I_GAS)||(i->signal==I_ACC)) {s->speed=1; s->mode=1; }
    else if(i->signal==I_BUTTON) s->button=0; 
  }
  else if((s->mode==1) && (s->speed==1) && (s->button==1)) {
    if(i->signal==I_GAS) { s->speed=2; s->mode=2; }
    else if(i->signal==I_BRAKE) { s->speed=0; s->mode=2; }
    else if(i->signal==I_BUTTON) { s->mode=0; s->button=0; }
  }
  else if((s->mode==2) && (s->speed==2) && (s->button==1)) {
    if((i->signal==I_BRAKE)||(i->signal==I_DEC)) { s->speed=1; s->mode=1; }
    else if(i->signal==I_BUTTON) { s->mode=0; s->button=0; }
  }
  else if((s->mode==2) && (s->speed==0) && (s->button==1)) {
    if((i->signal==I_GAS)||(i->signal==I_ACC)) { s->speed=1; s->mode=1; }
    else if(i->signal==I_BUTTON) { s->mode=0; s->button=0; }
  }
  else if((s->mode==0) && (s->speed==2) && (s->button==0)) {
    if((i->signal==I_BRAKE)||(i->signal==I_DEC)) s->speed=1; 
    else if(i->signal==I_BUTTON) s->button=1; 
  }
  else if((s->mode==0) && (s->speed==2) && (s->button==1)) {
    if((i->signal==I_BRAKE)||(i->signal==I_DEC)) { s->speed=1; s->mode=1; }
    else if(i->signal==I_BUTTON) s->button=0; 
  }
}

#ifdef CBMC
void input_assume(t_input* i) {
  __CPROVER_assume((1<=i->signal) && (i->signal<=5));
}
#else
#define __CPROVER_assert(a,b) if(a) { printf("%s passed\n",b); } else { printf("%s failed\n",b); }
#endif

#define HIST_LEN 5

extern t_input nondet_input();

int main(int argc, char** argv) {
  t_state s[HIST_LEN];
  t_input i[HIST_LEN];
  init(&s[HIST_LEN-1]);
  unsigned j=0;
  for(;j<HIST_LEN-1;j++) 
  {
    s[j] = s[HIST_LEN-1];
#ifdef CBMC
    input_assume(&i[j]);
#endif
  }
  
#ifdef CBMC
  while(1)
  {
#else 
  unsigned k=0;
  for(;k<TESTLENGTH;k++)
  {
    printf("iteration: %d\n",k);
#endif
    //save history
    unsigned j=1;
    for(;j<HIST_LEN;j++) s[j-1] = s[j];
    
    //step
#ifdef CBMC
    t_input new_i = nondet_input();
    input_assume(&new_i);
#else
    t_input new_i = TESTVECTOR[k];
#endif
    for(j=1;j<HIST_LEN;j++) i[j-1] = i[j];
    i[HIST_LEN-1] = new_i;
    compute(&new_i,&s[HIST_LEN-1]);

    //properties
    
    if((s[HIST_LEN-2].mode==1) && (s[HIST_LEN-2].speed==1) && 
       (s[HIST_LEN-2].button==1) && (i[HIST_LEN-1].signal==I_DEC))
      __CPROVER_assert(s[HIST_LEN-1].speed==1,"c::prop1");

    if((s[HIST_LEN-2].mode==2) && (s[HIST_LEN-2].speed==2) && 
       (i[HIST_LEN-1].signal==I_DEC))
      __CPROVER_assert(s[HIST_LEN-1].mode==1,"c::prop2");

    if((s[HIST_LEN-2].mode==1) && (i[HIST_LEN-1].signal==I_BRAKE))
       __CPROVER_assert(s[HIST_LEN-1].mode==2,"c::prop3");

    if((s[HIST_LEN-2].mode==0) && (s[HIST_LEN-2].speed==2) && 
        (s[HIST_LEN-2].button==0) && 
       (i[HIST_LEN-1].signal==I_BUTTON))
      __CPROVER_assert(s[HIST_LEN-1].button==1,"c::prop4");

    if((s[HIST_LEN-2].mode==0) && (s[HIST_LEN-2].button==0) && 
       (i[HIST_LEN-1].signal==I_BUTTON))
      __CPROVER_assert(s[HIST_LEN-1].button==1,"c::prop5");

    if((s[HIST_LEN-2].mode==1) && 
       (i[HIST_LEN-1].signal==I_GAS || i[HIST_LEN-1].signal==I_BRAKE))
      __CPROVER_assert(s[HIST_LEN-1].mode==2,"c::prop6");

    if((s[HIST_LEN-2].button==1) && (i[HIST_LEN-1].signal==I_BUTTON))
      __CPROVER_assert(s[HIST_LEN-1].button==0,"c::prop7");
 
    if((s[HIST_LEN-2].speed==2) && (i[HIST_LEN-1].signal==I_BRAKE))
      __CPROVER_assert(s[HIST_LEN-1].speed==1,"c::prop8");

    if((s[HIST_LEN-2].speed==1) && (i[HIST_LEN-1].signal==I_BRAKE))
      __CPROVER_assert(s[HIST_LEN-1].speed==0,"c::prop9");

    if((s[HIST_LEN-2].speed==0) && (i[HIST_LEN-1].signal==I_GAS))
      __CPROVER_assert(s[HIST_LEN-1].speed==1,"c::prop10");

    if((s[HIST_LEN-2].speed==1) && (i[HIST_LEN-1].signal==I_GAS))
      __CPROVER_assert(s[HIST_LEN-1].speed==2,"c::prop11");

    if((s[HIST_LEN-2].speed==2) && (i[HIST_LEN-1].signal==I_GAS))
      __CPROVER_assert(s[HIST_LEN-1].speed==2,"c::prop12");

    if((s[HIST_LEN-2].speed==0) && (i[HIST_LEN-1].signal==I_BRAKE))
       __CPROVER_assert(s[HIST_LEN-1].speed==0,"c::prop13");

    if((s[HIST_LEN-3].speed==1) && (i[HIST_LEN-2].signal==I_GAS) &&
       (i[HIST_LEN-1].signal==I_BRAKE))
      __CPROVER_assert(s[HIST_LEN-1].speed==1,"c::prop14");

    if((s[HIST_LEN-5].speed==0) && (i[HIST_LEN-4].signal==I_GAS) &&
        (i[HIST_LEN-3].signal==I_GAS) &&
        (i[HIST_LEN-2].signal==I_BRAKE) &&
        (i[HIST_LEN-1].signal==I_BRAKE))
      __CPROVER_assert(s[HIST_LEN-1].speed==0,"c::prop15");

    if((s[HIST_LEN-3].speed>=0) && (s[HIST_LEN-3].speed<=1) && 
        (i[HIST_LEN-2].signal==I_GAS) && (i[HIST_LEN-1].signal==I_BRAKE))
      __CPROVER_assert(s[HIST_LEN-1].speed==s[HIST_LEN-3].speed,"c::prop16");

    if(s[HIST_LEN-2].mode==0 && s[HIST_LEN-2].button==0 &&
      (i[HIST_LEN-1].signal!=I_BUTTON))
      __CPROVER_assert(s[HIST_LEN-1].mode==0,"c::prop18");

    if(s[HIST_LEN-3].mode==1 && s[HIST_LEN-3].speed==1 && 
        i[HIST_LEN-2].signal==I_BRAKE && i[HIST_LEN-1].signal==I_BRAKE)
      __CPROVER_assert(s[HIST_LEN-1].speed==0,"c::prop19");

    //final
    if((s[HIST_LEN-1].mode==0) && (s[HIST_LEN-1].speed==0) && 
       (s[HIST_LEN-1].button==0)) 
      __CPROVER_assert(1,"c::final");

  }

  return 0;
}
