/*******************************************************************************/
/* Cruise controller                                                           */
/* Author: Peter Schrammel                                                     */
/*******************************************************************************/
/* modelled after                                                              */
/*     Robert Bosch GmbH: Bosch Automotive Handbook. Bentley (2007)            */
/* if you reuse this code, please cite also                                    */
/*      Peter Schrammel, Tom Melham, Daniel Kroening. Chaining Test Cases for 
        Reactive System Testing. In Testing Software and Systems
        LNCS Volume 8254, 2013, pp 133-148                                     */
/*******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

typedef struct state {
  int mode;
  int speed;
  int button;
} t_state;

#define I_BUTTON 1
#define I_ACC 2
#define I_DEC 3
#define I_GAS 4
#define I_BRAKE 5

typedef struct input {
  int signal; 
} t_input;

void init(t_state *s) {
  s->mode = 0;
  s->speed = 0;
  s->button = 0;
}

void compute(t_input* i, t_state *s) {
  if((s->mode==0) && (s->speed==0) && (s->button==0)) {
    if((i->signal==I_ACC)||(i->signal==I_GAS)) s->speed=1;
    else if(i->signal==I_BUTTON) s->button=1;
  }
  else if((s->mode==0) && (s->speed==1) && (s->button==0)) {
    if((i->signal==I_BRAKE)||(i->signal==I_DEC)) s->speed=0;
    else if((i->signal==I_GAS)||(i->signal==I_ACC)) s->speed=2;
    else if(i->signal==I_BUTTON) { s->button=1; s->mode=1; }
  }
  else if((s->mode==0) && (s->speed==0) && (s->button==1)) {
    if((i->signal==I_GAS)||(i->signal==I_ACC)) {s->speed=1; s->mode=1; }
    else if(i->signal==I_BUTTON) s->button=0; 
  }
  else if((s->mode==1) && (s->speed==1) && (s->button==1)) {
    if(i->signal==I_GAS) { s->speed=2; s->mode=2; }
    else if(i->signal==I_BRAKE) { s->speed=0; s->mode=2; }
    else if(i->signal==I_BUTTON) { s->mode=0; s->button=0; }
  }
  else if((s->mode==2) && (s->speed==2) && (s->button==1)) {
    if((i->signal==I_BRAKE)||(i->signal==I_DEC)) { s->speed=1; s->mode=1; }
    else if(i->signal==I_BUTTON) { s->mode=0; s->button=0; }
  }
  else if((s->mode==2) && (s->speed==0) && (s->button==1)) {
    if((i->signal==I_GAS)||(i->signal==I_ACC)) { s->speed=1; s->mode=1; }
    else if(i->signal==I_BUTTON) { s->mode=0; s->button=0; }
  }
  else if((s->mode==0) && (s->speed==2) && (s->button==0)) {
    if((i->signal==I_BRAKE)||(i->signal==I_DEC)) s->speed=1; 
    else if(i->signal==I_BUTTON) s->button=1; 
  }
  else if((s->mode==0) && (s->speed==2) && (s->button==1)) {
    if((i->signal==I_BRAKE)||(i->signal==I_DEC)) { s->speed=1; s->mode=1; }
    else if(i->signal==I_BUTTON) s->button=0; 
  }
}

#ifndef NO_CBMC

void state_assume(t_state* s) {
  __CPROVER_assume(((s->mode==0) && ((s->speed==0) || (s->speed==2)) && 
                                    (0<=s->button) && (s->button<=1)) ||
                   ((s->mode==0) && (s->speed==1) && (s->button==0)) ||
                   ((s->mode==1) && (s->speed==1) && (s->button==1)) ||
		   ((s->mode==2) && ((s->speed==0) || (s->speed==2)) && 
                                    (s->button==1))
		  );
}

void input_assume(t_input* i, t_state* s) {
  __CPROVER_assume((1<=i->signal) && (i->signal<=5));
}

void final(t_input* i, t_state* s) {
  __CPROVER_assume((s->mode==0) && (s->speed==0) && (s->button==0));
}

void prop1(t_input *i, t_state* s) {
  //precondition
  __CPROVER_assume((s->mode==1) && (s->speed==1) && (s->button==1) && (i->signal==I_DEC));
  //step
  compute(i,s);
  //postcondition
  assert(s->speed==1);
}

void prop2(t_input *i, t_state* s) {
  //precondition
  __CPROVER_assume((s->mode==2) && (s->speed==2) && (i->signal==I_DEC));
  //step
  compute(i,s);
  //postcondition
  assert(s->mode==1);
}

void prop3(t_input *i, t_state* s) {
  //precondition
  __CPROVER_assume((s->mode==1) && (i->signal==I_BRAKE));
  //step
  compute(i,s);
  //postcondition
  assert(s->mode==2);
}

void prop4(t_input *i, t_state* s) {
  //precondition
  __CPROVER_assume((s->mode==0) && (s->speed==2) && (s->button==0) && (i->signal==I_BUTTON));
  //step
  compute(i,s);
  //postcondition
  assert(s->button==1);
}

void prop5(t_state* s,t_input *i) {
  //precondition
  __CPROVER_assume((s->mode==0) && (s->button==0) && (i->signal==I_BUTTON));
  //step
  compute(i,s);
  //postcondition
  assert(s->button==1);
}

void prop6(t_input *i, t_state* s) {
  //precondition
  __CPROVER_assume((s->mode==1) && (i->signal==I_GAS || i->signal==I_BRAKE));
  //step
  compute(i,s);
  //postcondition
  assert(s->mode==2);
}

void prop7(t_input *i, t_state* s) {
  //precondition
  __CPROVER_assume((s->button==1) && (i->signal==I_BUTTON));
  //step
  compute(i,s);
  //postcondition
  assert(s->button==0);
}

void prop8(t_input *i, t_state* s) {
  //precondition
  __CPROVER_assume((s->speed==2) && (i->signal==I_BRAKE));
  //step
  compute(i,s);
  //postcondition
  assert(s->speed==1);
}

void prop9(t_input *i, t_state* s) {
  //precondition
  __CPROVER_assume((s->speed==1) && (i->signal==I_BRAKE));
  //step
  compute(i,s);
  //postcondition
  assert(s->speed==0);
}

void prop10(t_input *i, t_state* s) {
  //precondition
  __CPROVER_assume((s->speed==0) && (i->signal==I_GAS));
  //step
  compute(i,s);
  //postcondition
  assert(s->speed==1);
}

void prop11(t_input *i, t_state* s) {
  //precondition
  __CPROVER_assume((s->speed==1) && (i->signal==I_GAS));
  //step
  compute(i,s);
  //postcondition
  assert(s->speed==2);
}

void prop12(t_input *i, t_state* s) {
  //precondition
  __CPROVER_assume((s->speed==2) && (i->signal==I_GAS));
  //step
  compute(i,s);
  //postcondition
  assert(s->speed==2);
}

void prop13(t_input *i, t_state* s) {
  //precondition
  __CPROVER_assume((s->speed==0) && (i->signal==I_BRAKE));
  //step
  compute(i,s);
  //postcondition
  assert(s->speed==0);
}

extern t_input nondet_input();

void prop14(t_input *i, t_state* s) {
  //precondition1
  __CPROVER_assume((s->speed==1) && (i->signal==I_GAS));
  //step
  compute(i,s);
  //precondition2
  *i = nondet_input();
  __CPROVER_assume(i->signal==I_BRAKE);
  //step
  compute(i,s);
  //postcondition
  assert(s->speed==1);
}

void prop15(t_input *i, t_state* s) {
  __CPROVER_assume((s->speed==0) && (i->signal==I_GAS));
  compute(i,s);
  *i = nondet_input();
  __CPROVER_assume(i->signal==I_GAS);
  compute(i,s);
  *i = nondet_input();
  __CPROVER_assume(i->signal==I_BRAKE);
  compute(i,s);
  *i = nondet_input();
  __CPROVER_assume(i->signal==I_BRAKE);
  compute(i,s);
  assert(s->speed==0);
}

void prop16(t_input *i, t_state* s) {
  int old_speed = s-> speed;
  __CPROVER_assume(i->signal==I_GAS);
  compute(i,s);
  *i = nondet_input();
  __CPROVER_assume(i->signal==I_BRAKE);
  compute(i,s);
  assert(s->speed==old_speed);
}

void prop17(t_input *i, t_state* s) {
  __CPROVER_assume(s->mode==1);
  assert(s->speed==1);
}

void prop18(t_input *i, t_state* s) {
  __CPROVER_assume(s->mode==0 && s->button==0);
  assert(s->speed==0);
}

void prop19(t_input* i1, t_state* s) {
  t_input i2;
  __CPROVER_assume(s->mode==1 && s->speed==1 &&
      i1->signal==4 && i2.signal==4);
  compute(i1,s);
  compute(&i2,s);
  assert(s->speed==0);
}

void main() {
}

#endif
