/*******************************************************************************/
/* Robotarm model                                                              */
/* Author: Peter Schrammel                                                     */
/*******************************************************************************/
/* if you reuse this code, please cite                                         */
/*      Peter Schrammel, Tom Melham, Daniel Kroening. Chaining Test Cases for 
        Reactive System Testing. In Testing Software and Systems
        LNCS Volume 8254, 2013, pp 133-148                                     */
/*******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

typedef struct state {
  int mode;
  int phase;
  int reset;
  double speed;
  double pos;
} t_state;

#define POS_INIT 0.0
#define POS_MAX 3.0
#define POS_MIN -3.0
#define A 0.25 //acceleration
#define V 0.5 //max speed

#define I_UP 1
#define I_NONE 0
#define I_DOWN -1

#define M_IDLE 0
#define M_UP 1
#define M_DOWN -1

#define P_CST 0
#define P_ACC 1
#define P_DEC -1

typedef struct io {
  int joystick; 
  int reset;
} t_io;

void init(t_state *s) {
  s->mode = M_IDLE;
  s->phase = P_CST;
  s->reset = 0;
  s->pos = POS_INIT;
  s->speed = 0.0;
}

void compute(t_io* i, t_state *s) {
  double abs_speed = s->speed;
  if(s->speed<0) abs_speed = -s->speed;
  double dist_est = A*(abs_speed/A)*(abs_speed/A)*1.0; //(safely?) estimate braking distance
  if(i->reset) { s->reset = 1; }
  if(s->mode==M_IDLE) {
    if(s->reset) {
      if(s->pos>POS_INIT) { s->mode = M_DOWN; s->phase = P_ACC; }
      else if(s->pos<POS_INIT) { s->mode = M_UP; s->phase = P_ACC; }
      else s->reset = 0;
    }
    else if(i->joystick==I_UP) {
      s->mode = M_UP;
      s->phase = P_ACC;
    }
    else if(i->joystick==I_DOWN) {
      s->mode = M_DOWN;
      s->phase = P_ACC;
    }
  }
  else if(s->mode==M_UP) {
    if(s->phase==P_CST) {
      if(!s->reset && (i->joystick==I_DOWN || s->pos>=POS_MAX-dist_est) ||
        s->reset && s->pos>=POS_INIT-dist_est) { s->phase = P_DEC; }
    }
    else if(s->phase==P_ACC) {
      if(!s->reset && (i->joystick==I_DOWN || s->pos>=POS_MAX-dist_est) ||
        s->reset && s->pos>=POS_INIT-dist_est) { s->phase = P_DEC; }
      else if(s->speed<V) s->speed += A;
      else { s->phase = P_CST; }
    }
    else if(s->phase==P_DEC) {
      if(s->speed>0) s->speed -= A;
      else { 
        if(!s->reset && i->joystick==I_DOWN) { s->mode = M_DOWN; s->phase = P_ACC; }
        else { s->mode = M_IDLE; s->phase = P_CST; }
      }
    }
  }
  else if(s->mode==M_DOWN) {
    if(s->phase==P_CST) {
      if(!s->reset && (i->joystick==I_UP || s->pos<=POS_MIN+dist_est) ||
        s->reset && s->pos<=POS_INIT+dist_est) { s->phase = P_DEC; }
    }
    else if(s->phase==P_ACC) {
      if(!s->reset && (i->joystick==I_UP || s->pos<=POS_MIN+dist_est) ||
        s->reset && s->pos<=POS_INIT+dist_est) { s->phase = P_DEC; }
      else if(s->speed>-V) s->speed -= A;
      else { s->phase = P_CST; }
    }
    else if(s->phase==P_DEC) {
      if(s->speed<0) s->speed += A;
      else { 
        if(i->joystick==I_UP) { s->mode = M_UP; s->phase = P_ACC; }
        else { s->mode = M_IDLE; s->phase = P_CST; }
      }
    }
  }
  s->pos += s->speed;
}

#ifndef NO_CBMC

void state_assume(t_state* s) {
  __CPROVER_assume(
    (-1<=s->mode && s->mode<=1)  &&
    (-1<=s->phase && s->phase<=1) &&
    (0<=s->reset && s->reset<=1) &&
    (s->mode==M_IDLE && s->phase==P_CST && s->speed==0.0 || !s->mode==M_IDLE)
  );
}

void input_assume(t_io* i) {
 __CPROVER_assume(
    -1<=i->joystick && i->joystick<=1 &&
    0<=i->reset && i->reset<=1
  );
}

void final(t_io* i, t_state* s) {
  __CPROVER_assume(s->mode==M_IDLE && !s->reset);
}

void prop1(t_io *i, t_state* s) {
  //precondition
  __CPROVER_assume(s->mode==M_IDLE && !s->reset && !i->reset && i->joystick==I_UP);
  //step
  compute(i,s);
  //postcondition
  assert(s->mode==M_UP && s->phase==P_ACC);
}

void prop2(t_io *i, t_state* s) {
  //precondition
  __CPROVER_assume(s->mode==M_IDLE && !s->reset && !i->reset && i->joystick==I_DOWN);
  //step
  compute(i,s);
  //postcondition
  assert(s->mode==M_DOWN && s->phase==P_ACC);
}

void prop3(t_io *i, t_state* s) {
  //precondition
  __CPROVER_assume(s->mode==M_IDLE && s->pos<POS_INIT && i->reset);
  //step
  compute(i,s);
  //postcondition
  assert(s->reset && s->mode==M_UP);
}

void prop4(t_io *i, t_state* s) {
  //precondition
  __CPROVER_assume(s->mode==M_IDLE && s->pos>POS_INIT && i->reset);
  //step
  compute(i,s);
  //postcondition
  assert(s->reset && s->mode==M_DOWN);
}

void prop5(t_io *i, t_state* s) {
  //precondition
  __CPROVER_assume(i->reset);
  //step
  compute(i,s);
  //postcondition
  assert(s->reset);
}

void prop6(t_io *i, t_state* s) {
  //precondition
  __CPROVER_assume(s->phase==P_ACC && (s->speed==V || s->speed==-V) &&
		   i->joystick==I_NONE && !i->reset);
  //step
  compute(i,s);
  //postcondition
  assert(s->phase==P_CST);
}

void prop7(t_io *i, t_state* s) {
  //precondition
  __CPROVER_assume(s->mode==M_UP && !s->reset && !i->reset && i->joystick==I_DOWN);
  //step
  compute(i,s);
  //postcondition
  assert(s->mode==M_UP && s->phase==P_DEC);
}

void prop8(t_io *i, t_state* s) {
  //precondition
  __CPROVER_assume(s->mode==M_DOWN && !s->reset && !i->reset && i->joystick==I_UP);
  //step
  compute(i,s);
  //postcondition
  assert(s->mode==M_DOWN && s->phase==P_DEC);
}

void prop9(t_io *i, t_state* s) {
  //precondition
  __CPROVER_assume(s->reset && (i->joystick==I_UP || i->joystick==I_DOWN));
  //step
  compute(i,s);
  //postcondition
  assert(s->reset);
}

void prop10(t_io *i, t_state* s) {
  //precondition
  __CPROVER_assume(s->mode==M_IDLE && !s->reset && !i->reset && (i->joystick==I_UP || i->joystick==I_DOWN));
  //step
  compute(i,s);
  //postcondition
  assert(s->mode!=M_IDLE && s->phase==P_ACC);
}

void prop11(t_io *i, t_state* s) {
  //precondition
  __CPROVER_assume(s->mode==M_UP && i->joystick==I_UP);
  //step
  compute(i,s);
  //postcondition
  assert(s->mode==M_UP);
}

void prop12(t_io *i, t_state* s) {
  //precondition
  __CPROVER_assume(s->mode==M_DOWN && i->joystick==I_DOWN);
  //step
  compute(i,s);
  //postcondition
  assert(s->mode==M_DOWN);
}

void prop13(t_io *i, t_state* s) {
  //precondition
  __CPROVER_assume(s->phase==P_DEC && s->speed==0.0 &&
		   i->joystick==I_NONE && !i->reset);
  //step
  compute(i,s);
  //postcondition
  assert(s->mode==M_IDLE && s->phase==P_CST);
}

void prop14(t_io *i, t_state* s) {
  //precondition
  __CPROVER_assume(s->mode==M_UP && s->phase==P_ACC && s->speed==V &&
		   i->joystick==I_NONE && !i->reset);
  //step
  compute(i,s);
  //postcondition
  assert(s->mode==M_UP && s->phase==P_CST && s->speed==V);
}

void prop15(t_io *i, t_state* s) {
  //precondition
  __CPROVER_assume(s->mode==M_DOWN && s->phase==P_ACC && s->speed==-V &&
		   i->joystick==I_NONE && !i->reset);
  //step
  compute(i,s);
  //postcondition
  assert(s->mode==M_DOWN && s->phase==P_CST && s->speed==-V);
}

void prop16(t_io *i, t_state* s) {
  //precondition
  __CPROVER_assume(s->mode==M_IDLE && i->joystick==I_NONE && !i->reset);
  //step
  compute(i,s);
  //postcondition
  assert(s->mode==M_IDLE && s->speed==0.0);
}

void prop17(t_io *i, t_state* s) {
  //precondition
  __CPROVER_assume(s->mode==M_IDLE && s->reset && s->pos==POS_INIT);
  //step
  compute(i,s);
  //postcondition
  assert(!s->reset);
}

void prop18(t_io *i, t_state* s) {
  //precondition
  __CPROVER_assume(s->pos==POS_MAX);
  //step
  compute(i,s);
  //postcondition
  assert(s->speed<=0);
}

void prop19(t_io *i, t_state* s) {
  //precondition
  __CPROVER_assume(s->pos==POS_MIN);
  //step
  compute(i,s);
  //postcondition
  assert(s->speed>=0);
}

#endif
